import os
import random
import json
from urllib.parse import urljoin
import uuid
import requests
import pandas as pd
from pydantic import BaseModel, ConfigDict
from dash import Dash, html, dcc, no_update, page_container, page_registry, register_page, Output, Input, State
from dash.exceptions import PreventUpdate
from dash.dependencies import ALL
import dash_bootstrap_components as dbc

from .world import World


API_URL = 'http://localhost:8000'


class Index(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str
    app: Dash
    collection: str = None
    columns: list = None
    title: str = None

    def __init__(self, app, collection=None, title=None, columns=None, **kwargs):
        super().__init__(
            id=str(uuid.uuid4()),
            app=app,
            collection=collection,
            columns=columns,
            title=title or collection,
            **kwargs,
        )

        @self.app.callback(
            Output(
                f'{self.id}-query-results', 'children',
            ),
            [
                Input(f'{self.id}-search-submit', 'n_clicks'),
            ],
            [
                State(f'{self.id}-search', 'value'),
            ],
        )
        def submit_form(n_clicks, value):
            if n_clicks is None or n_clicks == 0:
                return self.fetch()
            
            return self.fetch(query=value)

    def fetch(self, query=None):
        TEXT_MESSAGE_STYLE = {
            'width': '100%',
            'height': '100px',
            'padding': '2rem 0',
            'text-align': 'center',
        }

        if self.collection is None:
            return html.P('No collection specified.', style=TEXT_MESSAGE_STYLE)
        api_path = urljoin(API_URL, f'/{self.collection}')
        response = requests.get(api_path, params={'query': query})
        if response.status_code == 200:
            data = response.json()
            l = data.get('list', [])
        else:
            raise Exception(f'Error getting index {self.collection}: {response.status_code}')
        
        if len(l) == 0:
            return html.P('Nothing to see here.', style=TEXT_MESSAGE_STYLE),

        df = pd.DataFrame(l)
        df['id'] = df.apply(self.generate_link, axis=1)

        if self.columns is not None:
            df = df[self.columns]
        
        for col in df.columns:
            df[col] = df[col].apply(lambda x: x[:100] if type(x) == str else x)
        
        for col in df.columns:
            df[col] = df[col].apply(lambda x: json.dumps(x) if type(x) == dict else x)
        
        return dbc.Table.from_dataframe(df, style={
            'box-shadow': 'none',
            'border-top': '1px solid lightgray',
        })

    def generate_link(self, row):
        link = os.path.join(f'/{self.collection}', row['id'])
        return html.A(row.get('id'), href=link, target='_self')

    def load(self, data):
        if data is None:
            return
        name = data.get('name')
        if name is None:
            name = data.get('details', {}).get('name')
        self.collection = name
        print(f'Loaded index {self.collection}', data)

    def render(self, **kwargs):
        RESULTS_STYLE = {
        }

        results = html.Div([], id=f'{self.id}-query-results', style=RESULTS_STYLE)

        search = dbc.InputGroup([
            dbc.InputGroupText(
                html.I(className="bi bi-search"),
                style={
                    'color': 'lightgray',
                }
            ),
            dbc.Input(
                name='search',
                placeholder='Search',
                id=f'{self.id}-search',
            ),
        ])

        search_input = dbc.Form([
            dbc.Row([
                dbc.Col([
                    search,
                ], width=9),
                dbc.Col([
                    dbc.Button('Search', id=f'{self.id}-search-submit', n_clicks=0, color='secondary')
                ], width=3)
            ])
        ])

        HEADER_STYLE = {
            'padding': '10px',
        }

        INDEX_STYLE = {
            'background-color': 'white',
            'border-radius': '5px',
        }

        return html.Div(children=[
            dbc.Row(
                [
                    dbc.Col(
                        [
                            html.H4(self.title, style={
                                'font-size': '1.5rem',
                                'font-weight': 'normal',
                                'text-transform': 'capitalize',
                            }),
                        ],
                        width=3,
                    ),
                    dbc.Col(
                        [
                            html.Div(
                                search_input,
                            ),
                        ],
                        width=9,
                    ),
                ],
                style=HEADER_STYLE,
            ),
            dbc.Row(
                results,
            ),
            dcc.Interval(
                id=f'fetch-interval',
                n_intervals=1,  # Set an interval that triggers once
                max_intervals=1  # Only trigger once
            ),
            dcc.Store(id=f'{self.id}-data-store'),
        ], style=INDEX_STYLE, className='index')


class AdminPage(BaseModel):
    app: Dash
    name: str
    path: str = None
    path_template: str = None
    menu: bool = False
    icon: str = None
    api_path: str = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, app, name, path=None, path_template=None, menu=False, icon=None):
        super().__init__(
            app=app,
            name=name,
            path=path,
            path_template=path_template,
            menu=menu,
            icon=icon,
        )
        self.register_callbacks()

    def layout(self, **kwargs):
        content = self.render(**kwargs)
        return html.Div(children=[
            dcc.Interval(
                id='fetch-interval',
                interval=1,  # Set an interval that triggers once
                max_intervals=1  # Only trigger once
            ),
            content,
        ])
    
    def render(self, **kwargs):
        return html.Div(children=[
            html.H1(self.name),
        ])
    
    def register_callbacks(self):
        pass


class AdminIndexPage(AdminPage):
    index: Index = None
    collection: str = None

    def __init__(self, app, collection, columns=None, **kwargs):
        super().__init__(
            app,
            **kwargs,
        )
        self.collection = collection
        self.index = Index(
            app=self.app, collection=self.collection, columns=columns
        )

    def layout(self):
        content = self.index.render()
        return content


class EntityDetails(BaseModel):
    data: dict

    def render(self, **kwargs):
        def is_entity(v):
            if type(v) == dict:
                return 'id' in v and 'type' in v
            elif type(v) == list:
                return any([is_entity(item) for item in v])

        data = [
            {'field': k, 'value': v.get('title')} if type(v) == dict else {'field': k, 'value': v}
            for k, v in self.data.items()
            if k not in ['id', 'type', 'name', 'description'] and not is_entity(v)
        ]
        
        details = html.Div([
            html.H3(self.data.get('id')),
            html.P(self.data.get('type')),
            html.P(self.data.get('description')),
        ])

        list_items = dbc.ListGroup(
            [
                dbc.ListGroupItem(
                    [
                        html.H6(item['field']),
                        html.P(item['value'], style={
                            'max-width': '100%',
                            'max-height': '100px',
                            'overflow': 'hidden',
                        }),
                    ],
                )
                for item in data
            ],
        )

        return html.Div(
            [
                details,
                list_items,
            ],
            style={
                'padding': '1rem',
                'background-color': 'white',
            },
        )
    
    def load(self, data):
        pass

class EntityInputForm(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str
    name: str
    page: str
    app: Dash 
    data: dict = None

    def __init__(self, name, app, page, submit, **kwargs):
        super().__init__(
            id=str(uuid.uuid4()),
            name=name,
            app=app,
            page=page,
            **kwargs,
        )
        
        @self.app.callback(
            Output(
                f'{self.id}-form-output', 'children',
            ),
            [
                Input(f'{self.id}-submit', 'n_clicks'),
                Input(f'{self.page}-data-store', 'data'),
            ],
            [
                State({'type': 'form-input', 'field': ALL}, 'value')
            ]
        )
        def submit_form(n_clicks, data, values):
            if n_clicks is None or n_clicks == 0:
                raise PreventUpdate
            
            field_data = {}
            input_schema = json.loads(data.get('details', {}).get('input', '{}'))
            fields = input_schema.get('properties', {}).keys()
            for input_id, input_value in zip(fields, values):
                field_data[input_id] = input_value
            id = data.get('details', {}).get('id')
            return submit(id, field_data)

    def render(self):
        if self.data is None:
            return None
        input_data = self.data.get('input')
        if input_data is None:
            return None
        input = json.loads(input_data)
        if input is None or input == 'null':
            inputs = [{
                'id': 'input',
                'label': 'Input',
            }]
        else:
            input_schema = {
                name: field
                for name, field in input['properties'].items()
            }
            inputs = []
            for name, field in input_schema.items():
                input = {
                    'id': name,
                    'label': field['title'],
                }
                if field['type'] == 'string':
                    input['type'] = 'text'
                elif field['type'] == 'integer':
                    input['type'] = 'number'
                elif field['type'] == 'boolean':
                    input['type'] = 'checkbox'
                else:
                    input['type'] = 'text'
                
                inputs.append(input)
        
        form = dbc.Form(
            [
                html.H4(self.name, style={
                    'border-bottom': '1px solid lightgray',
                    'padding': '10px 0',
                    'font-weight': 'normal',
                }),
                *[
                    html.Div(
                        [
                            dbc.InputGroup([
                                dbc.InputGroupText(input['label']),
                                dbc.Input(
                                    id={'type': 'form-input', 'field': input['id']},
                                    name=input['label'],
                                    type=input['type'],
                                    placeholder=f'Enter {input["label"]}',
                                )
                            ]),
                        ],
                        style={
                            'margin': '0 0 10px 0',
                        }
                    )
                    for input in inputs
                ],
                html.Div(
                    dbc.Button('Submit', id=f'{self.id}-submit', n_clicks=0, color='secondary'),
                    style={
                        'margin': '0 0 10px 0',
                    }
                ),
                html.Div(id=f'{self.id}-form-output'),
            ],
            style={
                'padding': '10px',
                'margin': '10px 0',
                'background-color': 'white',
            }
        )
        return form
    
    def load(self, data):
        self.data = data


class AdminEntityPage(AdminPage):
    components: list = None

    def __init__(self, app, components=None, **kwargs):
        super().__init__(
            app,
            **kwargs,
        )
        self.components = components or []
    
    def register_callbacks(self):
        @self.app.callback(
            Output(f'{self.name}-data-store', 'data'),
            Output(f'{self.name}-details', 'children'),
            Output(f'{self.name}-components', 'children'),
            Input('fetch-interval', 'n_intervals'),
            Input('url', 'pathname'),
        )
        def fetch_data(n_intervals, pathname):
            if n_intervals is None or n_intervals == 0:
                return no_update
            api_path = urljoin(API_URL, pathname)
            response = requests.get(api_path)
            if response.status_code == 200:
                data = response.json()
                details = self.render_details(data)
                components = self.render_components(data)
                return data, details, components
            else:
                raise Exception(f'Error getting entity ({api_path}): {response.status_code}')

    def layout(self, **kwargs):
        return html.Div(children=[
            dcc.Location(id='url', refresh=False), 
            dcc.Loading(id='loading', type='default', children=[
                dcc.Store(id=f'{self.name}-data-store'),
                html.Div(id=f'{self.name}-details'),
                html.Div(id=f'{self.name}-components'),
            ]),
            
            dcc.Interval(
                id=f'fetch-interval',
                interval=1,  # Set an interval that triggers once
                max_intervals=1  # Only trigger once
            ),
        ])
    
    def render_details(self, data):
        details = data.get('details')
        if details is None:
            return None
        return EntityDetails(data=details).render()
    
    def render_components(self, data):
        for component in self.components:
            try:
                component.load(data)
            except Exception as e:
                print(f'Error loading component: {e}')

        details = data.get('details')
        if details is None:
            return None
        return [
            component.render()
            for component in self.components
        ]
    

class TemplateIndex(AdminIndexPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            'templates',
            name="Templates",
            path="/templates",
            columns=['id', 'name', 'instructions'],
            icon='bi bi-code-slash',
            **kwargs,
        )


class AgentIndex(AdminIndexPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Agents",
            path="/agents",
            collection='agents',
            icon='bi bi-robot',
            **kwargs,
        )


class QueryIndex(AdminIndexPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Queries",
            path="/queries",
            collection='queries',
            icon='bi bi-search',
            **kwargs,
        )


class SubscriptionIndex(AdminIndexPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Subscriptions",
            path="/subscriptions",
            collection='subscriptions',
            icon='bi bi-inbox',
            **kwargs,
        )


class ModelIndex(AdminIndexPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Models",
            path="/models",
            collection='models',
            icon='bi bi-infinity',
            **kwargs,
        )


class CollectionIndex(AdminIndexPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Collections",
            path="/collections",
            collection='collections',
            columns=['id', 'name', 'description'],
            icon='bi bi-database',
            **kwargs,
        )


class Inbox(AdminIndexPage):
    menu: bool = False

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Inbox",
            path="/inbox",
            collection='inbox',
            icon='bi bi-envelope',
            **kwargs,
        )


class Logs(AdminIndexPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Logs",
            path="/logs",
            collection='logs',
            icon='bi bi-list',
            **kwargs,
        )


class DetailsPage(AdminEntityPage):

    def __init__(self, app, collection, name=None, **kwargs):
        super().__init__(
            app,
            name=name or collection,
            path_template=f"/{collection}/<id>",
            **kwargs,
        )


class TemplateDetailsPage(AdminEntityPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Template Details",
            path_template="/templates/<id>",
            **kwargs,
        )
        results_index = Index(
            app=app, 
            collection='logs', 
            columns=['id', 'name', 'instructions']
        )

        def handle_submit(id, data):
            api_path = urljoin(API_URL, f'/templates/{id}/run')
            response = requests.post(api_path, json={'input': data})
            if response.status_code == 200:
                data = response.json()
                return data
            else:
                raise Exception(f'Error submitting form: {response.status_code}')

        input_form = EntityInputForm(
            name='Run Template',
            app=app,
            page=self.name,
            submit=handle_submit,
        )

        components = [
            input_form,
            results_index,
        ]

        self.components = components
    

class CollectionDetailsPage(AdminEntityPage):

    def __init__(self, app, **kwargs):
        super().__init__(
            app,
            name="Collection",
            path_template="/collections/<id>",
            **kwargs,
        )

        records = Index(
            app,
            title='Records',
        )

        self.components = [
            records,
        ]


class Admin:
    world: World

    def __init__(self, world):
        self.world = world
        self.app = Dash(
            world.name, 
            use_pages=True,
            pages_folder='',
            external_stylesheets=[dbc.themes.ZEPHYR, dbc.icons.FONT_AWESOME, dbc.icons.BOOTSTRAP],
        )

        pages = [
            DetailsPage(self.app, '<collection>', name='Entity Details'),
            DetailsPage(self.app, 'logs'),
            DetailsPage(self.app, 'queries'),
            TemplateDetailsPage(self.app),
            CollectionDetailsPage(self.app),

            Inbox(self.app, menu=False),
            QueryIndex(self.app, menu=False),
            SubscriptionIndex(self.app, menu=False),
            TemplateIndex(self.app, menu=True),
            CollectionIndex(self.app, menu=True),
            AgentIndex(self.app, menu=True),
            ModelIndex(self.app, menu=True),
            Logs(self.app, menu=True),
        ]

        for page in pages:
            register_page(page.name, layout=page.layout, path=page.path, path_template=page.path_template)
        
        SIDEBAR_STYLE = {
            "position": "fixed",
            "top": "68px",
            "left": 0,
            "bottom": 0,
            "width": "18rem",
            "padding": "1rem 1rem",
            "background-color": "#fff",
        }

        CONTENT_STYLE = {
            "margin-top": "66px",
            "padding": "1rem 2rem 0 20rem",
            "background-color": "#F5F5F5",
            "width": "100vw",
            "min-height": "100vh",
        }

        menu = [page for page in pages if page.menu]
        nav = dbc.Nav(
            [
                dbc.NavLink(
                    [
                        html.I(className=page.icon),
                        '   ',
                        page_registry[page.name]['name'], 
                    ],
                    href=page_registry[page.name]['relative_path'], 
                    active='exact',
                )
                for page in menu
            ],
            vertical=True,
            pills=True,
        )

        # list of input placeholders for the command prompt to cycle through
        # and help show users how to get the most out of it
        placeholders = [
            'Summarize the latest AI papers on arXiv',
        ]

        styles = """
        .index table {
            max-width: 100%;
        }

        .index table td {
            max-width: 300px;
            max-height: 100px;
            overflow: hidden;
        }
        """
        data_uri = f"data:text/css;charset=utf-8,{styles}"
        placeholder = random.choice(placeholders)
        self.app.layout = html.Div([
            dbc.Navbar(
                dbc.Container(
                    children=[
                        dbc.Col(
                            dbc.NavbarBrand(
                                "Promptz",
                                href="/",
                                className="ml-2",
                            ),
                            width=3,
                        ),
                        dbc.Col(
                            dbc.InputGroup([
                                dbc.InputGroupText(
                                    html.I(className="bi bi-chevron-right", style={
                                        'font-size': '1.2rem',
                                        'line-height': '1rem',
                                        'color': 'lightgray',
                                    }),
                                ),
                                dbc.Input(
                                    type="text",
                                    placeholder=placeholder,
                                ),
                            ]),
                            width=6,
                        ),
                        dbc.Col(
                            dbc.NavItem(
                                dbc.NavLink(html.I(className="bi bi-person-circle"), href="/")
                            ),
                            width=3,
                            style={
                                'text-align': 'right',
                                'font-size': '1.5em',
                            },
                        )
                    ],
                    style={
                        'max-width': '100vw',
                        'padding': '0 2rem',
                    }
                ),
                color="white",
                dark=False,
                fixed='top',
                style={
                    'border-bottom': '1px solid lightgray',
                }
            ),
            dbc.Container(
                [
                    dbc.Row(
                        [
                            dbc.Col(
                                [
                                    nav,
                                ],
                                width=3,
                                style=SIDEBAR_STYLE,
                            ),
                            dbc.Col(
                                [
                                    page_container
                                ], 
                                style=CONTENT_STYLE,
                                width=9,
                            ),
                        ],
                    ),
                ],
                fluid=True,
                style={
                },
            ),
            html.Link(rel="stylesheet", href=data_uri)
        ])



import os
from loguru import logger
from rich import pretty
from rich.logging import RichHandler
import openai

from .world import World
from .api import API
from .admin import Admin
from .collection import ChromaVectorDB
from .models.openai import ChatGPT


class App:
    name: str
    path: str
    world: World

    def __init__(self, name, path, world=None, db=None, llm=None, **kwargs):
        self.name = name
        self.path = path
        self.world = world or World(name, db=db, default_llm=llm)
    
    @classmethod
    def load(cls, path, db=None, llm=None, env=None):
        db = db or ChromaVectorDB(path=path)

        if llm is None:
            default_llm_id = os.environ.get('PXX_DEFAULT_LLM', 'default')

            if default_llm_id == 'chatgpt':
                api_key = os.environ.get('PXX_OPENAI_API_KEY')
                org_id = os.environ.get('PXX_OPENAI_ORG_ID')
                openai.api_key = api_key or os.environ.get('OPENAI_API_KEY')
                openai.organization = org_id or os.environ.get('OPENAI_ORG_ID')
                llm = llm or ChatGPT(id='default', api_key=env.get('PXX_OPENAI_API_KEY'), org_id=env.get('PXX_OPENAI_ORG_ID'))

        config = {
            'name': 'local',
            'path': path,
            'db': db,
            'llm': llm,
        }

        pretty.install()

        env = {**os.environ, **(env or {})}
        log_file_path = f"./log/{env.get('PXX_ENV', 'development')}.log"
        level = env.get('PXX_LOG_LEVEL', 'INFO')

        # Configure Loguru
        logger.remove()
        logger.add(
            log_file_path,
            rotation="10 MB",
            level=level,
            format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <level>{message}</level>",
            backtrace=True
        )

        logger.info("Log file: " + log_file_path)
        logger.info("Log level: " + level)

        return cls(**config)

    @property
    def api(self):
        if self._api is None:
            self._api = API(self.world)
        return self._api

    @property
    def admin(self):
        if self._admin is None:
            self._admin = Admin(self.world)
        return self._admin
    
    def __repr__(self):
        return f'<App {self.name} path={self.path}>'




import uuid
import json
from enum import Enum
from datetime import datetime
from abc import abstractmethod
from typing import *
from loguru import logger
import jsonschema
import pandas as pd
from pydantic import BaseModel, Field, ConfigDict, create_model
from pydantic_core._pydantic_core import PydanticUndefinedType
import chromadb


REGISTERED_ENTITIES = {}

JSON_TYPE_MAP: Dict[str, Type[Union[str, int, float, bool, Any]]] = {
    "string": str,
    "integer": int,
    "number": float,
    "boolean": bool,
    "object": dict,
    "array": list,
}

PYTYPE_TO_JSONTYPE: Dict[Type, str] = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    dict: "object",
    list: "array",
}


def _is_list(schema):
    return schema.get('type') == 'array'


def _is_list_type(type_hint):
    origin = get_origin(type_hint)
    return origin is list or (origin is List and len(get_args(type_hint)) == 1)


def _get_title(schema):
    return schema.get('title', schema.get('items', {}).get('title', 'Entity'))


def _get_properties(schema):
    properties = schema.get('properties', {}) if not _is_list(schema) else schema.get('items', {}).get('properties', {})
    return properties


def _get_field_type(field_info, definitions):
    field_type = field_info.get('type')
    if field_type is None:
        ref = field_info.get('$ref')
        if ref is None:
            ref = field_info.get('allOf', [{}])[0].get('$ref')
        if ref is None:
            return str
        ref_name = ref.split('/')[-1]
        field_type = ref_name
        definition = definitions.get(ref_name, {})
        if 'enum' in definition:
            members = {v: v for v in definition['enum']}
            E = Enum(definition.get('title', ref_name), members)
            return E
        else:
            M = create_model_from_schema(definition)
            return M

    if field_type == 'array':
        info = field_info.get('items', {})
        return List[_get_field_type(info, definitions)]
    return JSON_TYPE_MAP[field_type]


def _create_field(field_info, definitions, required=False):
    field_type = _get_field_type(field_info, definitions)
    field_default = field_info.get('default', ... if required else None)
    return (field_type, field_default)


def model_to_json_schema(model):
    """
    Convert a Pydantic BaseModel or Python data type to a JSON schema.

    Args:
        model: A Pydantic BaseModel, a Python data type, a list of BaseModel instances, or a dictionary.

    Returns:
        dict: A JSON schema representation of the input model.

    This function takes various types of input and converts them into a JSON schema representation:

    - If `model` is a Pydantic BaseModel, it extracts its schema using `model.schema()`.

    - If `model` is a Python data type (e.g., str, int, float), it maps it to the corresponding JSON type.

    - If `model` is a list of Pydantic BaseModels, it generates a JSON schema for an array of the BaseModel's schema.

    - If `model` is a dictionary, it is returned as is.

    Example:
    >>> from pydantic import BaseModel
    >>> class Person(BaseModel):
    ...     name: str
    ...     age: int
    ...
    >>> schema = model_to_json_schema(Person)
    >>> print(schema)
    {
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
            'age': {'type': 'integer'}
        },
        'required': ['name']
    }
    """
    output = None
    if isinstance(model, list):
        inner = model[0]
        if issubclass(inner, BaseModel):
            schema = inner.model_json_schema()
            output = {
                'type': 'array',
                'items': schema,
                '$defs': schema.get('$defs', {})
            }
        else:
            output = {
                'type': 'array',
                'items': {
                    'type': PYTYPE_TO_JSONTYPE[inner]
                }
            }
    elif isinstance(model, dict):
        output = model
    elif isinstance(model, BaseModel):
        output = model.model_json_schema()
    elif isinstance(model, type):
        if issubclass(model, Entity):
            output = model.model_json_schema()

            for name, field in model.__annotations__.items():
                if isinstance(field, type) and issubclass(field, Entity):
                    output['properties'][name] = {'$ref': '#/$defs/Query'}
                    output['$defs']['Query'] = {
                        'type': 'object',
                        'properties': {
                            'ids': {'type': 'array', 'items': {'type': 'string'}},
                            'query': { 'type': 'string' },
                            'collection': { 'type': 'string' },
                            'limit': { 'type': 'integer' },
                        },
                        'required': [],
                    }
                elif getattr(field, '__origin__', None) == list and isinstance(field.__args__[0], type) and issubclass(field.__args__[0], Entity):
                    output['properties'][name] = {'$ref': '#/$defs/Query'}
                    output['$defs']['Query'] = {
                        'type': 'object',
                        'properties': {
                            'ids': {'type': 'array', 'items': {'type': 'string'}},
                            'query': { 'type': 'string' },
                            'collection': { 'type': 'string' },
                            'limit': { 'type': 'integer' },
                        },
                        'required': [],
                    }
        elif issubclass(model, BaseModel):
            output = model.model_json_schema()
    
    return output


def create_model_from_schema(schema, base=None):
    """
    Create a Pydantic BaseModel from a JSON schema.

    Args:
        schema (dict): The JSON schema to create the Pydantic model from.

    Returns:
        pydantic.BaseModel: A Pydantic data model class generated from the schema.

    This function takes a JSON schema and generates a Pydantic BaseModel class
    with fields corresponding to the properties defined in the schema. It
    also handles definitions and required fields.

    If the schema doesn't specify a 'type' field, it defaults to 'Entity'.

    Example:
    >>> schema = {
    ...     'title': 'Person',
    ...     'type': 'object',
    ...     'properties': {
    ...         'name': {'type': 'string'},
    ...         'age': {'type': 'integer'}
    ...     },
    ...     'required': ['name']
    ... }
    >>> Person = create_model_from_schema(schema)
    >>> person = Person(name='Alice', age=30)
    >>> person.name
    'Alice'
    >>> person.age
    30
    """
    properties = _get_properties(schema)
    definitions = schema.get('$defs', {})
    required = schema.get('required', [])
    fields = {
        name: _create_field(field_info, definitions, name in required)
        for name, field_info in properties.items()
    }
    if 'id' not in fields:
        fields['id'] = (str, None)
    if 'type' not in fields:
        fields['type'] = (str, schema.get('title', 'Entity').lower())
    return create_model(schema.get('title', 'Entity').capitalize(), **fields, __base__=base)


def create_entity_from_schema(schema, data, session=None, base=None):
    """
    Create a Pydantic data entity from a JSON schema and input data.

    Args:
        schema (dict): The JSON schema that defines the structure of the entity.
        data (dict or list): The input data to populate the entity. For a single entity, provide a dictionary.
                             For a list of entities, provide a list of dictionaries.

    Returns:
        pydantic.BaseModel or List[pydantic.BaseModel]: A Pydantic data entity or a list of entities generated
                                                      from the schema and input data.

    This function takes a JSON schema and input data and creates a Pydantic data entity or a list of entities
    based on the schema and data provided. It handles properties, definitions, and optional fields defined
    in the schema.

    If the schema defines an entity as a list, the input data should be a list of dictionaries. Each dictionary
    represents an entity. If 'id' is not provided for each entity, it will be generated using a random UUID.

    If the schema defines an entity as an object (not a list), the input data should be a dictionary representing
    a single entity. If 'id' is not provided, it will be generated using a random UUID.

    Example:
    >>> schema = {
    ...     'title': 'Person',
    ...     'type': 'object',
    ...     'properties': {
    ...         'name': {'type': 'string'},
    ...         'age': {'type': 'integer'}
    ...     },
    ...     'required': ['name']
    ... }
    >>> data = {'name': 'Alice', 'age': 30}
    >>> person = create_entity_from_schema(session, schema, data)
    >>> person.name
    'Alice'
    >>> person.age
    30

    >>> schema_list = {
    ...     'title': 'People',
    ...     'type': 'array',
    ...     'items': {
    ...         'type': 'object',
    ...         'properties': {
    ...             'name': {'type': 'string'},
    ...             'age': {'type': 'integer'}
    ...         },
    ...         'required': ['name']
    ...     }
    ... }
    >>> data_list = [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
    >>> people = create_entity_from_schema(session, schema_list, data_list)
    >>> len(people)
    2
    >>> people[0].name
    'Alice'
    >>> people[1].age
    25
    """

    if _is_list(schema):
        data = [
            {**o, 'id': str(uuid.uuid4()) if o.get('id') is None else o['id']}
            for o in data
        ]
    else:
        if data.get('id') is None:
            data['id'] = str(uuid.uuid4())
    
    definitions = schema.get('definitions', {})
    for name, field in schema.get('properties', {}).items():
        _type = _get_field_type(field, definitions)
        if isinstance(_type, type) and issubclass(_type, Enum):
            if data.get(name):
                data[name] = data[name].name.lower()
        elif getattr(_type, '__origin__', None) == list and isinstance(_type.__args__[0], type) and issubclass(_type.__args__[0], Enum):
            if data.get(name):
                data[name] = [d.name.lower() for d in data[name]]
    
    def _field_serializer(obj):
        if isinstance(obj, Enum):
            return obj.value
        elif hasattr(obj, 'model_dump'):
            return {k: v for k, v in obj.model_dump().items() if v is not None}
        elif isinstance(obj, list):
            if all(hasattr(o, 'model_dump') for o in obj):
                return [{k: v for k, v in e.model_dump().items() if v is not None} for e in obj]
        return obj
    
    if _is_list(schema):
        data = [{k: _field_serializer(v) for k, v in item.items()} for item in data]
    else:
        data = {k: _field_serializer(v) for k, v in data.items()}

    jsonschema.validate(data, schema)
    m = create_model_from_schema(schema, base=base)
    defaults = {
        'type': _get_title(schema).lower(),
    }
    if _is_list(schema):
        return [m.load(session=session, **{**defaults, **o}) for o in data]
    else:
        return m.load(session=session, **{**defaults, **data})


def serializer(obj):
    if isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, BaseModel):
        return obj.model_json_schema()
    raise TypeError(f"Type {type(obj)} not serializable")


class Entity(BaseModel):
    id: str = None
    type: str = None
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    def __init__(self, **data):
        if 'type' not in data:
            data['type'] = self.__class__.__name__.lower()
        if 'id' not in data:
            data['id'] = str(uuid.uuid4())

        for k, v in data.items():
            if isinstance(v, Enum):
                data[k] = v.value
            elif isinstance(v, list) and all(isinstance(e, Enum) for e in v):
                data[k] = [e.value for e in v]
        
        super().__init__(**data)
    
    @classmethod
    def load(cls, session=None, **kwargs):
        for name, field in cls.__annotations__.items():
            if field.__name__ == 'Entity':
                def loader(self, cls=cls, session=session, name=name, field=field, data=kwargs):
                    # Lazy-loading logic here
                    logger.info(f'Loading {name}')
                    field_data = data.get(name)
                    if field_data is None:
                        return None
                    limit = field_data.get('limit')
                    collection = field_data.get('collection')
                    response = session.query(ids=field_data.get('ids'), limit=limit, collection=collection)
                    if response is None:
                        return None
                    if limit is None or limit > 1:
                        return response.objects
                    elif limit == 1:
                        return response.first
                    return None
                
                setattr(cls, name, property(loader))
            
        return cls(**kwargs)
    
    @classmethod
    def generate_schema_for_field(cls, name, field_type: Any, field: Field):
        return_list = False
        definitions = {}

        if _is_list_type(field_type):
            field_type = get_args(field_type)[0]
            return_list = True
        
        # Handle enums
        if isinstance(field_type, type) and issubclass(field_type, Enum):
            schema = {
                "type": "string",
                "enum": [e.name.lower() for e in field_type],
            }

        # Handle basic types
        elif isinstance(field_type, type) and issubclass(field_type, (int, float, str, bool)):
            type_ = PYTYPE_TO_JSONTYPE[field_type]
            schema = {"type": type_}

        # Handle Pydantic model types (reference schema)
        elif isinstance(field_type, type) and issubclass(field_type, Entity):
            schema = {
                "type": "object",
                "properties": {
                    "id": {"type": "string"},
                    "type": {"type": "string"},
                },
            }
        
        # Handle default case by getting the cls field and calling schema
        else:
            if isinstance(field_type, list):
                field_type = field_type[0]
            if field_type.__name__ not in definitions:
                definitions[field_type.__name__] = {
                    "type": "object",
                    "properties": field.default.schema() if field.default is not None else {}
                }
            schema = {"$ref": f"#/$defs/{field_type.__name__}"} 

        if return_list:
            schema = {
                "type": "array",
                "items": schema,
            }
        
        # TODO: need to test/fix this
        info = None
        if info is not None:
            if info.description:
                schema['description'] = info.description
            if info.ge:
                schema['ge'] = info.ge
            if info.gt:
                schema['gt'] = info.gt
            if info.le:
                schema['le'] = info.le
            if info.lt:
                schema['lt'] = info.lt
            if info.min_length:
                schema['min_length'] = info.min_length
            if info.max_length:
                schema['max_length'] = info.max_length
            
            extra = info.extra
            if extra is not None:
                if 'generate' in extra:
                    schema['generate'] = extra['generate']

        if field.default:
            schema['default'] = field.default
        return schema, definitions, []
    
    @classmethod
    def __schema(cls, by_alias: bool = True, **kwargs):
        properties = {}
        required = []
        definitions = {}

        for field_name, field in cls.model_fields.items():
            try:
                type_ = cls.__annotations__.get(field_name, field.annotation)
                field_schema, defs, reqs = cls.generate_schema_for_field(field_name, type_, field)
                properties[field_name] = field_schema
                definitions = {**definitions, **defs}
            except Exception as e:
                logger.error('schema field failed', field_name, e, field)
                continue
            
            if field.is_required():
                required.append(field_name)
            required += reqs

        # Construct the base schema
        base_schema = {
            "title": cls.__name__,
            "type": "object",
            "properties": properties,
            "$defs": definitions,  # Include definitions for references
            "required": required,
        }

        return base_schema
    
    def display(self):
        return self.model_dump_json()
    

class Query(BaseModel):
    type: str = 'query'
    query: str = None
    where: Dict[str, (int|str|bool)] = None
    collection: str = None

    def __init__(self, query=None, where=None, collection=None, **kwargs):
        super().__init__(query=query, where=where, collection=collection, **kwargs)


class Subscription(Entity):
    type: str = 'subscription'
    query: Query = None

    def __init__(self, query=None, **kwargs):
        super().__init__(query=query, **kwargs)


class VectorCollection:

    @abstractmethod
    def query(self, texts, where=None, ids=None, **kwargs):
        '''
        Return a list of results.
        '''
    
    @abstractmethod
    def get(self, ids=None, where=None, **kwargs):
        '''
        Return a list of results.
        '''
    
    @abstractmethod
    def upsert(self, ids, documents, metadatas, **kwargs):
        '''
        Upsert a list of results.
        '''
    
    @abstractmethod
    def delete(self, ids=None, where=None, **kwargs):
        '''
        Delete a list of results.
        '''


class VectorDB:
    name: str

    @abstractmethod
    def get_collection(self, name, **kwargs):
        '''
        Return a collection if it exists.
        '''

    @abstractmethod
    def get_or_create_collection(self, name, **kwargs):
        '''
        Return a collection or create a new one if it doesn't exist.
        '''

    @abstractmethod
    def delete_collection(self, name, **kwargs):
        '''
        Return a collection or create a new one if it doesn't exist.
        '''
    
    @abstractmethod
    def collections():
        '''
        Return a list of collections.
        '''


class ChromaVectorDB(VectorDB):

    def __init__(self, endpoint=None, api_key=None, path=None, **kwargs):
        self.client = chromadb.PersistentClient(path=f'{path}/.px/db' if path else "./.px/db")
    
    def get_or_create_collection(self, name, **kwargs):
        return self.client.get_or_create_collection(name, **kwargs)
    
    def create_collection(self, name, **kwargs):
        return self.client.create_collection(name, **kwargs)
    
    def get_collection(self, name, **kwargs):
        try:
            return self.client.get_collection(name, **kwargs)
        except ValueError:
            return None
    
    def delete_collection(self, name, **kwargs):
        return self.client.delete_collection(name, **kwargs)
    
    def collections(self):
        return self.client.list_collections()


class MemoryDB(VectorCollection):
    name: str

    def __init__(self, name, **kwargs):
        self.name = name

    def get(self, ids=None, where=None, **kwargs):
        return { 'ids': [], 'documents': [], 'metadatas': [] }
    
    def query(self, texts, where=None, ids=None, **kwargs):
        return []
    
    def upsert(self, ids, documents, metadatas, **kwargs):
        pass

    def delete(self, ids=None, where=None, **kwargs):
        pass


class MemoryVectorDB(VectorDB):
    
    def get_or_create_collection(self, name, **kwargs):
        return MemoryDB(name)
    
    def create_collection(self, name, **kwargs):
        pass
    
    def get_collection(self, name, **kwargs):
        return MemoryDB(name)
    
    def delete_collection(self, name, **kwargs):
        pass
    
    def collections(self):
        return []
    
    def upsert(self, ids, documents, metadatas, **kwargs):
        pass


class EntitySeries(pd.Series):

    @property
    def _constructor(self):
        return EntitySeries

    @property
    def _constructor_expanddim(self):
        return Collection
    
    @property
    def object(self):
        d = self.to_dict()
        return Entity(**d)


class Collection(pd.DataFrame):
    _metadata = ['db', 'schema', 'session']

    @property
    def _constructor(self, *args, **kwargs):
        return Collection
    
    @property
    def _constructor_sliced(self):
        return EntitySeries
    
    @classmethod
    def load(cls, session, db):
        records = db.get(where={'item': 1})
        docs = [
            {
                'id': id, 
                **json.loads(r), 
            } 
            for id, r, m in zip(records['ids'], records['documents'], records['metadatas'])
        ]
        c = Collection(docs)
        c.db = db
        c.session = session
        return c
    
    def embedding_query(self, *texts, ids=None, where=None, threshold=0.1, limit=None, **kwargs):
        texts = [t for t in texts if t is not None]
        
        scores = {}
        if len(texts) == 0:
            results = self.db.get(ids=ids, where=where, **kwargs)
            for id, m in zip(results['ids'], results['metadatas']):
                if m.get('item') != 1:
                    id = m.get('item_id')
                if id not in scores:
                    scores[id] = 1
                else:
                    scores[id] += 1
        else:
            results = self.db.query(query_texts=texts, where=where, **kwargs)
            for i in range(len(results['ids'])):
                for id, d, m in zip(results['ids'][i], results['distances'][i], results['metadatas'][i]):
                    if m.get('item') != 1:
                        id = m.get('item_id')
                    if id not in scores:
                        scores[id] = 1 - d
                    else:
                        scores[id] += 1 - d
        
        try:
            filtered_scores = {k: v for k, v in scores.items() if v >= threshold}
            sorted_ids = sorted(filtered_scores, key=filtered_scores.get, reverse=True)
            results = self[self['id'].isin(sorted_ids)].set_index('id').loc[sorted_ids].reset_index()
            logger.info(f'Found {len(results)} results for query: {texts}')
            if limit is not None:
                return results.head(limit)
            else:
                return results
        except KeyError as e:
            logger.error(f'Failed to parse query results: {e}')
            return None
    
    def __call__(self, *texts, where=None, **kwargs) -> Any:
        return self.embedding_query(*texts, where=where, **kwargs)
    
    @property
    def name(self):
        return self.db.name
    
    @property
    def objects(self):
        if self.empty:
            return []
        if hasattr(self, 'db'):
            ids = self['id'].values.tolist()
            d = self.db.get(ids=ids)
            m = {id: metadata for id, metadata in zip(d['ids'], d['metadatas'])}
            schemas = {
                id: json.loads(metadata['schema']) for id, metadata in m.items()
                    if 'schema' in metadata and metadata['schema'] is not None
            }
            return [
                create_entity_from_schema(
                    schemas.get(r['id']),
                    {
                        k: v for k, v in r.items() if (len(v) if isinstance(v, list) else pd.notnull(v))
                    },
                    session=self.session,
                    base=REGISTERED_ENTITIES.get(r['type'], Entity),
                ) 
                for r in self.to_dict('records')
            ]
        else:
            return [
                create_entity_from_schema(
                    self.schema or {},
                    {
                        k: v for k, v in r.items() if (len(v) if isinstance(v, list) else pd.notnull(v))
                    },
                    base=REGISTERED_ENTITIES.get(r.get('type'), Entity),
                ) 
                for r in self.to_dict('records')
            ]
    
    @property
    def first(self):
        objects = self.objects
        if len(objects) == 0:
            return None
        else:
            return objects[0]
    
    def delete(self, *items):
        self.db.delete(ids=[i.id.replace(' ', '') for i in items])
        for item in items:
            self.db.delete(where={'item_id': item.id})
        self.drop(self[self['id'].isin([i.id for i in items])].index, inplace=True)
        logger.info(f'Deleted {len(items)} items from {self.name}')

    def embed(self, *items, **kwargs):
        records = self._create_records(*items, **kwargs)
        if len(records) == 0:
            return None
        
        # dedupe the records based on id
        records = {r['id']: r for r in records}.values()

        ids = [r['id'] for r in records]
        self.db.upsert(
            ids=ids,
            documents=[r['document'] for r in records],
            metadatas=[r['metadata'] for r in records],
        )

        if self.empty:
            new_items = [r for r in records if r['metadata']['item'] == 1]
        else:
            new_items = [
                r for r in records 
                if r['id'] not in self['id'].values
                and r['metadata']['item'] == 1
            ]
        docs = [{'id': r['id'], **json.loads(r['document'])} for r in new_items]
        df = pd.concat([self, Collection(docs)], ignore_index=True)
        self.drop(self.index, inplace=True)
        for column in df.columns:
            self[column] = df[column]
        logger.info(f'Embedded {len(new_items)} items into {self.name}')

        for item in items:
            if item.type not in REGISTERED_ENTITIES:
                REGISTERED_ENTITIES[item.type.lower()] = item.__class__
        return self

    def _create_records(self, *items, **kwargs):
        records = []

        def _field_serializer(obj):
            if isinstance(obj, Enum):
                return obj.value
            elif isinstance(obj, Entity):
                return {
                    'ids': [obj.id],
                    'collection': self.name,
                    'limit': 1,
                }
            elif isinstance(obj, list):
                if len(obj) and isinstance(obj[0], Entity):
                    return {
                        'ids': [o.id for o in obj],
                        'collection': self.name,
                    }
            raise TypeError(f"Type {type(obj)} not serializable")

        def _serializer(obj):
            if isinstance(obj, Enum):
                return obj.value
            elif isinstance(obj, Entity):
                record = { 'id': obj.id, 'type': obj.type }
                for name, field in obj.model_dump().items():
                    if name in ['id', 'type']:
                        continue
                    f = obj.model_fields.get(name)
                    if f is None:
                        logger.error(f'Field {name} not found in {obj.__class__}')
                        continue
                    if isinstance(f.annotation, type) and issubclass(f.annotation, Entity):
                        field = _field_serializer(getattr(obj, name))
                    if f.json_schema_extra and f.json_schema_extra.get('embed', True) == False:
                        continue
                    record[name] = field
            raise TypeError(f"Type {type(obj)} is not serializable")

        def _schema_serializer(obj):
            if isinstance(obj, Enum):
                return obj.value
            elif isinstance(obj, PydanticUndefinedType):
                return None
            elif isinstance(obj, BaseModel):
                return obj.model_json_schema()
            raise TypeError(f"Type {type(obj)} not serializable", obj)

        for item in items:
            now = datetime.now().isoformat()
            
            if isinstance(item, str):
                item = Entity(type='string', value=item)
            
            for name, field in item.model_dump().items():
                if name in ['id', 'type']:
                    continue

                f = item.model_fields.get(name)
                if f is None:
                    continue

                document = json.dumps({name: field}, default=_serializer)
                if isinstance(f.annotation, type) and issubclass(f.annotation, Entity):
                    # TODO: Handle nested fields
                    logger.debug(f'Field {name} is an Entity')
                    continue
                if f.json_schema_extra and f.json_schema_extra.get('embed', True) == False:
                    continue
                if isinstance(field, int) or isinstance(field, float) or isinstance(field, bool):
                    continue

                field_record = {
                    'id': f'{item.id}_{name}',
                    'document': document,
                    'metadata': {
                        'field': name,
                        'collection': self.name,
                        'item': 0,
                        'item_id': item.id,
                        'created_at': now,
                    },
                }
                records.append(field_record)

            doc = { k: v for k, v in item.model_dump().items() if k not in ['id'] }
            for k, v in doc.items():
                # if v represents an Entity field then create records for it and replace it with a reference
                # use model_fields to check if the current v is an Entity
                f = item.model_fields.get(k)
                if f is None:
                    continue
                if v is None:
                    continue
                if isinstance(f.annotation, type) and issubclass(f.annotation, Entity):
                    doc[k] = _field_serializer(getattr(item, k))
                    records += self._create_records(getattr(item, k))
                elif getattr(f.annotation, '__origin__', None) == list and isinstance(f.annotation.__args__[0], type) and issubclass(f.annotation.__args__[0], Entity):
                    doc[k] = _field_serializer(getattr(item, k))
                    records += self._create_records(*getattr(item, k))

            doc_record = {
                'id': item.id,
                'document': json.dumps(doc, default=_serializer),
                'metadata': {
                    'collection': self.name,
                    'type': item.type,
                    'item': 1,
                    'schema': json.dumps(model_to_json_schema(item.__class__)),
                    'created_at': now,
                },
            }
            records.append(doc_record)
        
        return records


class CollectionEntity(Entity):
    type: str = 'collection'
    name: str = None
    description: str = None

    def __init__(self, name, description=None, records=None, **kwargs):
        super().__init__(
            name=name, description=description, records=records, **kwargs
        )


import json
import uuid
import numpy as np
from loguru import logger
from pydantic import BaseModel
from typing import Any, Dict, List

from .collection import Collection, CollectionEntity, Query, VectorDB, model_to_json_schema
from .template import Template, TemplateRunner, MaxRetriesExceeded, MockLLM
from .models import PromptLog, QueryLog, LLM
from .models.openai import ChatGPT


class Session:

    def __init__(self, world, default_collection='default'):
        self.world = world
        self._collection = default_collection
    
    def _run_prompt(self, t, input, llm, context=None, history=None, dryrun=False, retries=3, to_json=False, allow_none=False, **kwargs):
        e = None
        s = self.world.template_system
        rendered = s.render(t, {'input': s.parse(input)})
        try:
            r = s(self, t, input, context=context, llm=llm, history=history, dryrun=dryrun, retries=retries, allow_none=allow_none,**kwargs)
            log = PromptLog(template=t.id, raw_input=rendered, raw_output=r.raw)
            self.store(log, collection='logs')
            if isinstance(r.content, list):
                if len(r.content) > 0 and isinstance(r.content[0], BaseModel):
                    es = [dict(e) for e in r.content]
                else:
                    es = r.content
                if to_json:
                    return es
                else:
                    c = Collection(es)
                    if t.output is not None:
                        output_schema = json.loads(t.output)
                        # if output schema is a list, get the inner type
                        if output_schema.get('type', None) == 'array':
                            output_schema = output_schema.get('items', {})
                        c.schema = output_schema
                    return c
            else:
                if to_json:
                    return dict(r.content)
                else:
                    return r.content
        except MaxRetriesExceeded as e:
            logger.error(f'Max retries exceeded: {e}')
            log = PromptLog(template=t.id, raw_input=rendered, error=str(e))
            self.store(log, collection='logs')
            raise e
    
    def prompt(self, instructions=None, input=None, output=None, id=None, context=None, template=None, llm=None, examples=None, allow_none=False, logs=None, tools=None, dryrun=False, retries=3, debug=False, silent=False, to_json=False, **kwargs):
        if output is not None:
            output = model_to_json_schema(output)
            if output is not None:
                output = json.dumps(output)

        if template is None:
            tvars = dict(
                id=id,
                output=output,
                instructions=instructions,
                context=context,
                examples=examples,
                logs=logs,
                tools=tools,
                debug=debug,
                silent=silent,
            )
            template = Template(**{k: v for k, v in tvars.items() if v is not None})
        elif isinstance(template, str):
            template = self.query(ids=[template], collection='templates').first
        
        if template is None:
            raise ValueError(f'No template found with id {template}')
        
        if llm is None:
            response = self.query(ids=['default'], collection='models')
            if response is not None:
                llm = response.first
        elif isinstance(llm, str):
            response = self.query(ids=[llm], collection='models')
            if response is not None:
                llm = response.first
        
        if llm is None and self.world.default_llm is not None:
            llm = self.world.default_llm
        
        if llm is None:
            raise ValueError(f'No model found')

        return self._run_prompt(template, input, llm, dryrun=dryrun, retries=retries, to_json=to_json, allow_none=allow_none, **kwargs)
    
    def embed(self, item, field=None):
        if isinstance(item, str):
            return self.ef([item])[0]
        elif isinstance(item, BaseModel):
            return self.ef([item.model_dump_json()])[0]
        elif isinstance(item, list):
            return [self.embed(i, field=field) for i in item]
    
    def query(self, *texts, field=None, ids=None, where=None, collection=None, limit=None):
        c = self.collection(collection)
        if c is None:
            raise ValueError(f'No collection found with name {collection}')
        where = where or {}
        if field is not None:
            where['field'] = field
        r =  c(*texts, ids=ids, where=where, limit=limit)
        if r is None:
            return None

        def serialize(item):
            if isinstance(item, BaseModel):
                return item.model_dump_json()
            else:
                return item
        
        if False:
            # TODO: fix this
            log = QueryLog(
                query=texts,
                where=where,
                collection=collection,
                result=json.dumps(r.objects, default=serialize),
            )
            self.store(log, collection='queries')
        return r
    
    def chat(self, message, context=None, agent=None, **kwargs):
        raise NotImplementedError()

    def store(self, *items, collection=None):
        def flatten(lst):
            return [
                item for sublist in lst 
                for item in (
                    flatten(sublist) if isinstance(sublist, list) else [sublist]
                )
            ]
        
        items = flatten([
            item.objects if isinstance(item, Collection) else item 
            for item in items
        ])
        
        c = self.collection(collection)
        if c is None:
            self.create_collection(collection)
            c = self.collection(collection)
        c.embed(*[item for item in items if item is not None])
    
    def delete(self, *items, collection=None):
        c = self.collection(collection)
        if c is None:
            raise ValueError(f'No collection found with name {collection}')
        
        c.delete(*items)
    
    def delete_collection(self, collection):
        self.world.delete_collection(collection)
    
    def create_collection(self, collection):
        self.world.create_collection(self, collection)
    
    def collections(self):
        return self.world.db.collections()
    
    def chain(self, *steps, llm=None, **kwargs):
        llm = llm or self.llm
        input = None
        cc = self.collection(f'chain.{uuid.uuid4()}')
        for step in steps:
            if isinstance(step, Template):
                p = step
                p.llm = llm
                input = self._run_prompt(p, input, **kwargs)
                self.store(*input, collection=cc.name)
            if isinstance(step, str):
                p = Template(step, llm=llm)
                input = self._run_prompt(p, input, **kwargs)
                if isinstance(input, list):
                    if isinstance(input[0], BaseModel):
                        self.store(*input, collection=cc.name)
                elif isinstance(input, BaseModel):
                    self.store(input, collection=cc.name)
            if isinstance(step, dict):
                input = {
                    k: self.chain(v, llm=llm, **kwargs)
                    for k, v in step.items()
                }
            if isinstance(step, list):
                input = [
                    self.chain(v, llm=llm, **kwargs)
                    for v in step
                ]
            if isinstance(step, Query):
                qc = self.collection(step.collection) if step.collection is not None else cc
                input = qc(step.query or '', where=step.where)
        return cc
        
    def collection(self, name=None):
        if name is None:
            name = self._collection
        collection = self.world._collections.get(name)
        return collection
    
    def evaluate(self, actual, expected, **kwargs):
        threshold = 0.69
        ex = self.embed(actual)
        ey = self.embed(expected)
        dot_product = np.dot(ex, ey)
        norm_1 = np.linalg.norm(ex)
        norm_2 = np.linalg.norm(ey)

        cosine_similarity = dot_product / (norm_1 * norm_2)
        if cosine_similarity < threshold:
            cosine_similarity = 0
        rescaled = (cosine_similarity - threshold) * (100 / (1 - threshold))
        return max(0, rescaled)
    
    def __call__(self, collection: Collection, *args, **kwargs):
        return self.run(collection, *self._processors, **kwargs)
    
    @property
    def logs(self):
        return self.collection('logs')


class World:
    name: str
    sessions: List[Session]
    _collections: Dict[str, Collection]
    db: VectorDB
    default_llm: LLM

    def __init__(self, name, db, default_llm=None):
        self.name = name
        self.sessions = []
        self._collections = {}
        # TODO: hack, should register as normal system
        self.template_system = TemplateRunner()
        self.db = db
        self.default_llm = default_llm
        
        session = self.create_session('setup')
        collection = self.db.get_or_create_collection('collections')
        self.collections = Collection.load(session, collection)
        self.create_collection(session, 'default', 'Default collection used when calling store()')
        self.create_collection(session, 'logs', 'Stores a log of all prompts and their outputs')
        self.create_collection(session, 'queries', 'Query stored objects in collections')
        self.create_collection(session, 'subscriptions', 'Subscriptions to queries')
        self.create_collection(session, 'agents', 'Configurations for interactive and autonomous AI agents')
        self.create_collection(session, 'templates', 'Prompt templates used to interact with AI models')
        self.create_collection(session, 'models', 'Configurations for AI models')
        
        for collection in self.db.collections():
            self.create_collection(session, collection.name)

        session.store(self.default_llm, collection='models')

    def create_session(self, name=None, user=None):
        logger.info(f'Creating session: {name} / {user}')
        session = Session(self)
        self.sessions.append(session)
        return session
    
    def create_collection(self, session, name, description=None, metadata=None):
        collection = self.db.get_collection(name)
        if collection is None:
            if metadata is None:
                metadata = {"hnsw:space": "cosine"}
            collection = self.db.create_collection(name, metadata=metadata)
            r = CollectionEntity(name=name, description=description or '')
            self.collections.embed(r)
        c = Collection.load(session, collection)
        self._collections[name] = c
        return collection
    
    def delete_collection(self, name):
        self.db.delete_collection(name)
        if name in self._collections:
            del self._collections[name]

    @property
    def templates(self):
        return self._collections['templates']
    
    @property
    def logs(self):
        return self._collections['logs']

    def __call__(self, session, *args: Any, **kwds: Any) -> Any:
        for system in self.systems().objects:
            q = Query(**system.query)
            items = session.query(q.query, where=q.where, collection=q.collection)
            try:
                _system = self._systems[system.name]
                updates = _system(items)
                if updates is not None:
                    session.store(updates)
            except Exception as e:
                raise e


import random
import uuid
import pytest

from . import User, Trait, Account, _user, user, session


def test_store(session, user):
    session.store(user)
    x = session.query(ids=[user.id]).first

    assert x is not None
    assert x.id == user.id
    assert x.name == user.name
    assert x.age == user.age
    assert x.role.value == user.role.value
    assert x.banned == user.banned
    assert x.vigor == user.vigor

def test_store__multiple(session):
    n = 2
    ids = [str(uuid.uuid4()) for _ in range(n)]
    traits = random.choices(list(Trait), k=3)
    session.store(
        *[
            User(id=ids[i], name="test", age=20, traits=traits)
            for i in range(n)
        ]
    )

    users = session.query().objects
    stored_ids = [user.id for user in users]
    assert all([id in stored_ids for id in ids])

def test_store__alt_collection(session, user):
    session.store(user, collection='alt-test')

    assert session.query(ids=[user.id]) is None

    user = session.query(ids=[user.id], collection='alt-test').first
    assert user is not None

def test_store__list_field(session, user):
    traits = random.choices(list(Trait), k=3)
    user.traits = traits
    session.store(user)
    x = session.query(ids=[user.id]).first

    assert x is not None
    assert len(x.traits) == len(traits)
    assert x.traits[0].value == traits[0].value
    assert x.traits[1].value == traits[1].value
    assert x.traits[2].value == traits[2].value

def test_query__ids(session):
    users = [_user() for _ in range(3)]
    session.store(*users)
    x = session.query(ids=[u.id for u in users])

    assert x is not None
    assert len(x.objects) == len(users)
    assert x[x['id'] == users[0].id].first.id == users[0].id

def test_foriegn_key(session, user):
    account = Account(user=user)
    session.store(account)

    _user = session.query(ids=[user.id]).first
    assert _user is not None
    assert _user.id == user.id
    assert _user.name == user.name

    _account = session.query(ids=[account.id]).first
    assert _account is not None
    assert _account.id == account.id
    assert _account.user.id == user.id
    assert _account.user.name == user.name

def test_one_to_many(session, user):
    payees = [_user() for _ in range(3)]
    account = Account(user=user, payees=payees)
    session.store(account)

    for user in session.query(ids=[u.id for u in payees]).objects:
        assert user is not None
        assert user.id in [u.id for u in payees]

    _account = session.query(ids=[account.id]).first
    assert _account is not None
    assert len(_account.payees) == len(payees)
    assert _account.payees[0].id == payees[0].id
    assert _account.payees[0].name == payees[0].name
    assert _account.payees[1].id == payees[1].id


import uuid
from enum import Enum
from typing import *
import jsonschema
from pydantic import BaseModel, Field, ConfigDict, create_model
from IPython.display import display, HTML


JSON_TYPE_MAP: Dict[str, Type[Union[str, int, float, bool, Any]]] = {
    "string": str,
    "integer": int,
    "number": float,
    "boolean": bool,
    "object": dict,
    "array": list,
}

PYTYPE_TO_JSONTYPE: Dict[Type, str] = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    dict: "object",
    list: "array",
}


def _is_list(schema):
    return schema.get('type') == 'array'


def _is_list_type(type_hint):
    origin = get_origin(type_hint)
    return origin is list or (origin is List and len(get_args(type_hint)) == 1)


def _get_title(schema):
    return schema.get('title', schema.get('items', {}).get('title', 'Entity'))


def _get_properties(schema):
    properties = schema.get('properties', {}) if not _is_list(schema) else schema.get('items', {}).get('properties', {})
    return properties


def _get_field_type(field_info, definitions):
    field_type = field_info.get('type')
    if field_type is None:
        ref = field_info.get('$ref')
        if ref is None:
            ref = field_info.get('allOf', [{}])[0].get('$ref')
        if ref is None:
            return str
        ref_name = ref.split('/')[-1]
        field_type = ref_name
        definition = definitions.get(ref_name, {})
        if 'enum' in definition:
            members = {v: v for v in definition['enum']}
            E = Enum(definition.get('title', ref_name), members)
            return E
        else:
            M = create_model_from_schema(definition)
            return M

    if field_type == 'array':
        info = field_info.get('items', {})
        return List[_get_field_type(info, definitions)]
    return JSON_TYPE_MAP[field_type]


def _create_field(field_info, definitions, required=False):
    field_type = _get_field_type(field_info, definitions)
    field_default = field_info.get('default', ... if required else None)
    return (field_type, field_default)


def model_to_json_schema(model):
    """
    Convert a Pydantic BaseModel or Python data type to a JSON schema.

    Args:
        model: A Pydantic BaseModel, a Python data type, a list of BaseModel instances, or a dictionary.

    Returns:
        dict: A JSON schema representation of the input model.

    This function takes various types of input and converts them into a JSON schema representation:

    - If `model` is a Pydantic BaseModel, it extracts its schema using `model.schema()`.

    - If `model` is a Python data type (e.g., str, int, float), it maps it to the corresponding JSON type.

    - If `model` is a list of Pydantic BaseModels, it generates a JSON schema for an array of the BaseModel's schema.

    - If `model` is a dictionary, it is returned as is.

    Example:
    >>> from pydantic import BaseModel
    >>> class Person(BaseModel):
    ...     name: str
    ...     age: int
    ...
    >>> schema = model_to_json_schema(Person)
    >>> print(schema)
    {
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
            'age': {'type': 'integer'}
        },
        'required': ['name']
    }
    """
    output = None
    if isinstance(model, list):
        inner = model[0]
        if issubclass(inner, BaseModel):
            schema = inner.model_json_schema()
            output = {
                'type': 'array',
                'items': schema,
                '$defs': schema.get('$defs', {})
            }
        else:
            output = {
                'type': 'array',
                'items': {
                    'type': PYTYPE_TO_JSONTYPE[inner]
                }
            }
    elif isinstance(model, dict):
        output = model
    elif isinstance(model, BaseModel):
        output = model.model_json_schema()
    elif isinstance(model, type):
        if issubclass(model, BaseModel):
            output = model.model_json_schema()
    
    return output


def create_model_from_schema(schema, base=None):
    """
    Create a Pydantic BaseModel from a JSON schema.

    Args:
        schema (dict): The JSON schema to create the Pydantic model from.

    Returns:
        pydantic.BaseModel: A Pydantic data model class generated from the schema.

    This function takes a JSON schema and generates a Pydantic BaseModel class
    with fields corresponding to the properties defined in the schema. It
    also handles definitions and required fields.

    If the schema doesn't specify a 'type' field, it defaults to 'Entity'.

    Example:
    >>> schema = {
    ...     'title': 'Person',
    ...     'type': 'object',
    ...     'properties': {
    ...         'name': {'type': 'string'},
    ...         'age': {'type': 'integer'}
    ...     },
    ...     'required': ['name']
    ... }
    >>> Person = create_model_from_schema(schema)
    >>> person = Person(name='Alice', age=30)
    >>> person.name
    'Alice'
    >>> person.age
    30
    """
    properties = _get_properties(schema)
    definitions = schema.get('$defs', {})
    required = schema.get('required', [])
    fields = {
        name: _create_field(field_info, definitions, name in required)
        for name, field_info in properties.items()
    }
    if 'id' not in fields:
        fields['id'] = (str, None)
    if 'type' not in fields:
        fields['type'] = (str, schema.get('title', 'Entity').lower())
    return create_model(schema.get('title', 'Entity').capitalize(), **fields, __base__=base)


def create_entity_from_schema(schema, data, base=None):
    """
    Create a Pydantic data entity from a JSON schema and input data.

    Args:
        schema (dict): The JSON schema that defines the structure of the entity.
        data (dict or list): The input data to populate the entity. For a single entity, provide a dictionary.
                             For a list of entities, provide a list of dictionaries.

    Returns:
        pydantic.BaseModel or List[pydantic.BaseModel]: A Pydantic data entity or a list of entities generated
                                                      from the schema and input data.

    This function takes a JSON schema and input data and creates a Pydantic data entity or a list of entities
    based on the schema and data provided. It handles properties, definitions, and optional fields defined
    in the schema.

    If the schema defines an entity as a list, the input data should be a list of dictionaries. Each dictionary
    represents an entity. If 'id' is not provided for each entity, it will be generated using a random UUID.

    If the schema defines an entity as an object (not a list), the input data should be a dictionary representing
    a single entity. If 'id' is not provided, it will be generated using a random UUID.

    Example:
    >>> schema = {
    ...     'title': 'Person',
    ...     'type': 'object',
    ...     'properties': {
    ...         'name': {'type': 'string'},
    ...         'age': {'type': 'integer'}
    ...     },
    ...     'required': ['name']
    ... }
    >>> data = {'name': 'Alice', 'age': 30}
    >>> person = create_entity_from_schema(schema, data)
    >>> person.name
    'Alice'
    >>> person.age
    30

    >>> schema_list = {
    ...     'title': 'People',
    ...     'type': 'array',
    ...     'items': {
    ...         'type': 'object',
    ...         'properties': {
    ...             'name': {'type': 'string'},
    ...             'age': {'type': 'integer'}
    ...         },
    ...         'required': ['name']
    ...     }
    ... }
    >>> data_list = [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
    >>> people = create_entity_from_schema(schema_list, data_list)
    >>> len(people)
    2
    >>> people[0].name
    'Alice'
    >>> people[1].age
    25
    """

    if _is_list(schema):
        data = [
            {**o, 'id': str(uuid.uuid4()) if o.get('id') is None else o['id']}
            for o in data
        ]
    else:
        if data.get('id') is None:
            data['id'] = str(uuid.uuid4())
    
    definitions = schema.get('definitions', {})
    for name, field in schema.get('properties', {}).items():
        _type = _get_field_type(field, definitions)
        if isinstance(_type, type) and issubclass(_type, Enum):
            if data.get(name):
                data[name] = data[name].name.lower()
        elif getattr(_type, '__origin__', None) == list and isinstance(_type.__args__[0], type) and issubclass(_type.__args__[0], Enum):
            if data.get(name):
                data[name] = [d.name.lower() for d in data[name]]
    
    # TODO: need to somehow handle nested entities which should be stored as IDs
    # currently the schema is correct in that it defines the desired type as a string
    # however, the entity needs to be loaded when the parent entity is loaded
    
    jsonschema.validate(data, schema)
    m = create_model_from_schema(schema, base=base)
    defaults = {
        'type': _get_title(schema).lower(),
    }
    if _is_list(schema):
        return [m.load(**{**defaults, **o}) for o in data]
    else:
        return m.load(**{**defaults, **data})


def serializer(obj):
    if isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, BaseModel):
        return obj.model_json_schema()
    raise TypeError(f"Type {type(obj)} not serializable")


import pytest
from enum import Enum
from pydantic import BaseModel

from promptx.collection import *
from . import User, Account, _user, user, session


def test_objects_are_returned_with_correct_schema(session, mocker):
    user = User(name="test", age=20, traits=['nice', 'mean'])
    db = mocker.Mock(spec=VectorCollection)
    db.name = 'test'
    db.get.return_value = {'ids': [user.id], 'documents': [user.model_dump_json()], 'metadatas': [{'schema': json.dumps(user.model_json_schema())}]}
    collection = Collection.load(session, db)
    collection.embed(user)

    assert len(collection.objects) == 1
    assert collection.first.name == "test"

def test_embedding_an_entity(session, mocker):
    user = User(name="test", age=20, traits=['nice', 'mean'])
    db = mocker.Mock(spec=VectorCollection)
    db.name = 'test'
    db.get.return_value = {'ids': [user.id], 'documents': [user.model_dump_json()], 'metadatas': [{'schema': json.dumps(user.model_json_schema())}]}
    collection = Collection.load(session, db)
    collection.embed(user)

    assert len(collection.objects) == 1
    assert collection.first.name == "test"



def test_convert_pydantic_model_to_json_schema():
    schema = model_to_json_schema(User)
    expected_schema = {
        'title': 'User',
        'type': 'object',
        'properties': {
            'id': {'title': 'Id', 'type': 'string', 'default': None},
            'type': {'title': 'Type', 'type': 'string', 'default': None},
            'name': {'maxLength': 20, 'minLength': 3, 'title': 'Name', 'type': 'string'},
            'age': {'exclusiveMaximum': 100, 'minimum': 18, 'title': 'Age', 'type': 'integer'},
            'role': {'allOf': [{'$ref': '#/$defs/Role'}], 'default': 'admin'},
            'banned': {'default': False, 'generate': False, 'title': 'Banned', 'type': 'boolean'}, 
            'vigor': {'default': 0, 'maximum': 1.0, 'minimum': 0.0, 'title': 'Vigor', 'type': 'number'}, 
            'traits': {
                'title': 'Traits', 
                'type': 'array',
                'items': {'$ref': '#/$defs/Trait'}, 
                'maxItems': 3, 'minItems': 1, 
            },
        },
        '$defs': {
            'Role': {'enum': ['admin', 'user'], 'title': 'Role', 'type': 'string'}, 
            'Trait': {'enum': ['nice', 'mean', 'funny', 'smart'], 'title': 'Trait', 'type': 'string'},
        },
        'required': ['name', 'age', 'traits']
    }
    assert schema == expected_schema


def test_convert_list_of_pydantic_models_to_json_schema():
    class Item(BaseModel):
        name: str
        price: float
    schema = model_to_json_schema([Item])
    expected_schema = {
        'type': 'array',
        'items': {
            'title': 'Item',
            'type': 'object',
            'properties': {
                'name': {'title': 'Name', 'type': 'string'},
                'price': {'title': 'Price', 'type': 'number'}
            },
            'required': ['name', 'price']
        },
        '$defs': {},
    }
    assert schema == expected_schema


def test_passthrough_dictionary_as_json_schema():
    schema = {'type': 'object', 'properties': {'key': {'type': 'string'}}}
    result_schema = model_to_json_schema(schema)
    assert result_schema == schema


def test_convert_pydantic_model_with_nested_models_to_json_schema():
    class Address(BaseModel):
        street: str
        city: str

    class Person(BaseModel):
        name: str
        age: int
        address: Address

    schema = model_to_json_schema(Person)
    expected_schema = {
        'type': 'object',
        'title': 'Person',
        'properties': {
            'name': {'title': 'Name', 'type': 'string'},
            'age': {'title': 'Age', 'type': 'integer'},
            'address': {
                '$ref': '#/$defs/Address'
            }
        },
        'required': ['name', 'age', 'address'],
        '$defs': {
            'Address': {
                'type': 'object',
                'title': 'Address',
                'properties': {
                    'street': {'title': 'Street', 'type': 'string'},
                    'city': {'title': 'City', 'type': 'string'}
                },
                'required': ['street', 'city']
            }
        }
    }
    assert schema == expected_schema


def test_passthrough_dictionary_with_nested_schema_as_json_schema():
    schema = {
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
            'address': {
                'type': 'object',
                'properties': {
                    'street': {'type': 'string'},
                    'city': {'type': 'string'}
                },
                'required': ['street', 'city']
            }
        },
        'required': ['name'],
    }
    result_schema = model_to_json_schema(schema)
    assert result_schema == schema


def test_convert_pydantic_model_with_nested_entities_to_json_schema():
    schema = model_to_json_schema(Account)
    expected_schema = {
        'title': 'Account',
        'type': 'object',
        'properties': { 'id': {'title': 'Id', 'type': 'string', 'default': None},
            'type': {'title': 'Type', 'type': 'string', 'default': None},
            'user': { '$ref': '#/$defs/Query' },
            'payees': { '$ref': '#/$defs/Query' },
        },
        '$defs': {
            'Query': {
                'type': 'object',
                'properties': {
                    'ids': { 'type': 'array', 'items': { 'type': 'string' } },
                    'query': { 'type': 'string' },
                    'collection': { 'type': 'string' },
                    'limit': { 'type': 'integer' },
                },
                'required': []
            }
        },
        'required': ['user'],
    }
    assert schema['$defs']['Query'] == expected_schema['$defs']['Query']


def test_create_model_from_schema_with_required_fields():
    schema = {
        'title': 'Person',
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
            'age': {'type': 'integer'}
        },
        'required': ['name']
    }
    Person = create_model_from_schema(schema, base=Entity)
    person = Person(name='Alice', age=30)
    assert isinstance(person, BaseModel)
    assert person.name == 'Alice'
    assert person.age == 30


def test_create_model_from_schema_with_missing_optional_fields():
    # Create a schema without the 'age' property, which is optional
    schema = {
        'title': 'Person',
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
            'age': {'type': 'integer'}
        },
        'required': ['name']
    }
    Person = create_model_from_schema(schema, base=Entity)
    person = Person(name='Alice')
    assert isinstance(person, BaseModel)
    assert person.name == 'Alice'
    assert person.age is None


def test_create_model_from_schema_with_custom_type():
    # Create a schema with a custom 'type' field
    schema = {
        'title': 'Item',
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
            'price': {'type': 'number'}
        }
    }
    Item = create_model_from_schema(schema, base=Entity)
    item = Item(name='Product', price=10.99)
    assert isinstance(item, BaseModel)
    assert item.name == 'Product'
    assert item.price == 10.99


def test_create_model_from_schema_with_enum():
    # Example JSON schema with complex properties
    complex_schema = {
        'title': 'Person',
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
            'age': {'type': 'integer'},
            'gender': {
                'type': 'string',
                'enum': ['Male', 'Female', 'Other']
            },
            'address': {
                'type': 'object',
                'properties': {
                    'street': {'type': 'string'},
                    'city': {'type': 'string'}
                }
            },
            'hobbies': {
                'type': 'array',
                'items': {'type': 'string'}
            }
        },
        'required': ['name', 'age']
    }
    
    Person = create_model_from_schema(complex_schema, base=Entity)
    person = Person(
        name='Alice',
        age=30,
        gender='female',
        address={'street': '123 Main St', 'city': 'City'},
        hobbies=['Reading', 'Cooking']
    )
    assert person.name == 'Alice'
    assert person.age == 30
    assert person.gender == 'female' 
    assert person.address['street'] == '123 Main St'
    assert person.address['city'] == 'City'
    assert person.hobbies == ['Reading', 'Cooking']


# Example JSON schema for testing
person_schema = {
    'title': 'Person',
    'type': 'object',
    'properties': {
        'name': {'type': 'string'},
        'age': {'type': 'integer'}
    },
    'required': ['name']
}

def test_create_entity_from_schema_with_single_entity(session):
    data = {'name': 'Alice', 'age': 30}
    person = create_entity_from_schema(person_schema, data, session=session, base=Entity)
    assert isinstance(person, BaseModel)
    assert person.name == 'Alice'
    assert person.age == 30

def test_create_entity_from_schema_with_single_entity_missing_required_field(session):
    # Missing 'name' field, which is required
    data = {'age': 30}
    with pytest.raises(jsonschema.ValidationError):
        create_entity_from_schema(person_schema, data, session=session, base=Entity)

def test_create_entity_from_schema_with_single_entity_and_generated_id_and_type(session):
    data = {'name': 'Alice', 'age': 30}
    person = create_entity_from_schema(person_schema, data, session=session, base=Entity)
    assert hasattr(person, 'id')
    assert hasattr(person, 'type')
    assert person.type == 'person'

def test_create_entity_from_schema_with_list_of_entities(session):
    data_list = [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
    people_schema = {
        'type': 'array',
        'items': person_schema
    }
    people = create_entity_from_schema(people_schema, data_list, session=session, base=Entity)
    assert isinstance(people, list)
    assert len(people) == 2
    assert all(isinstance(person, BaseModel) for person in people)
    assert people[0].name == 'Alice'
    assert people[1].age == 25

def test_create_entity_from_schema_with_list_of_entities_and_generated_ids(session):
    data_list = [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
    people_schema = {
        'type': 'array',
        'items': person_schema
    }
    people = create_entity_from_schema(people_schema, data_list, session=session, base=Entity)
    assert all(hasattr(person, 'id') for person in people)  # 'id' should be generated for each entity



from .collection import Collection, Entity
from .template import Template


class DefaultChatTemplate(Template):

    template: str = """
    {{input}}
    Message: {{message}}
    """

    input_templat: str  = """
    Context: {{input}}
    """

    output_templat: str  = """
    {{output}}
    """


class ChatBot(Entity):
    name: str
    template: Template
    history: Collection

    def __init__(self, name, template=None, history=None, **kwargs):
        if history is None:
            history = Collection()
        
        if template is None:
            template = DefaultChatTemplate()

        super().__init__(name=name, template=template, history=history, **kwargs)


import os
import time
from typing import Callable, List
from loguru import logger

from .collection import Collection, MemoryVectorDB
from .world import Session
from .application import App
from .auth import DefaultUser 


def prompt(instructions=None, input=None, output=None, prompt=None, context=None, template=None, llm=None, examples=None, allow_none=False, history=None, tools=None, dryrun=False, retries=3, debug=False, silent=False, **kwargs):
    kwargs = dict(
        instructions=instructions,
        input=input,
        output=output,
        prompt=prompt,
        llm=llm,
        context=context,
        examples=examples,
        allow_none=allow_none,
        history=history,
        tools=tools,
        template=template,
        debug=debug,
        dryrun=dryrun,
        silent=silent,
        retries=retries,
    )
    return DEFAULT_SESSION.prompt(**kwargs)


def store(*items, collection=None, **kwargs) -> Collection:
    return DEFAULT_SESSION.store(*items, collection=collection, **kwargs)


def delete(*items, collection=None, **kwargs) -> None:
    return DEFAULT_SESSION.delete(*items, collection=collection, **kwargs)


def collections():
    return DEFAULT_SESSION.collections()


def delete_collection(collection, **kwargs) -> None:
    return DEFAULT_SESSION.delete_collection(collection, **kwargs)


def create_collection(collection, **kwargs) -> None:
    return DEFAULT_SESSION.create_collection(collection, **kwargs)


def query(*texts, field=None, where=None, collection=None, **kwargs) -> Collection:
    return DEFAULT_SESSION.query(
        *texts, field=field, where=where, collection=collection, **kwargs)


def chat(message, context=None, **kwargs):
    return DEFAULT_SESSION.chat(message, context, **kwargs)


def evaluate(*testcases, **kwargs) -> Collection:
    return DEFAULT_SESSION.evaluate(*testcases, **kwargs)


def session() -> Session:
    return DEFAULT_SESSION


DEFAULT_SESSION = None
def set_default_session(session):
    global DEFAULT_SESSION
    DEFAULT_SESSION = session


def find_project_root(path=None, config_filename='.px'):
    home_dir = os.path.expanduser("~")
    if path is None:
        path = os.getcwd()
    while path != home_dir:
        if os.path.exists(os.path.join(path, config_filename)) and os.path.isdir(path):
            return path
        path = os.path.dirname(path)
    if os.path.exists(os.path.join(home_dir, config_filename)):
        return home_dir
    return None


def load(path='local', **env):
    if path == 'local':
        project_path = find_project_root()
        if project_path is not None:
            path = project_path

    app = None
    if path == 'local':
        db = MemoryVectorDB()
        env = {**os.environ, **(env or {})}
        app = App.load(path, db=db, env=env)
    elif path.startswith('http'):
        print('loading remote app from', path)
        raise NotImplementedError
    else:
        logger.info(f'loading local app from {path}')
        from dotenv import load_dotenv
        load_dotenv(os.path.join(path, '.env'), override=True)
        logger.info(f'loaded environment variables from {os.path.join(path, ".env")}')
        logger.info(f'API KEY {os.environ.get("PXX_OPENAI_API_KEY")[-5:]}')
        app = App.load(path, env=env)

        # look for a config.py file and execute it
        config_file = os.path.join(path, 'config.py')
        if os.path.exists(config_file):
            import importlib.util
            spec = importlib.util.spec_from_file_location('config', config_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            session = app.world.create_session(f'{app.name}_config_{int(time.time())}')
            module.setup(session)
    
    user = DefaultUser()
    s = app.world.create_session(user)
    set_default_session(s)
    return app


import os
if os.environ.get('PXX_AUTOLOAD', True):
    load()


from typing import *

from .collection import Collection, VectorDB


class Query:
    pass


class System:
    pass


# Should space just be the nested space/subspace structure?
class Space:
    name: str
    _db: VectorDB

    def __init__(self, name):
        super().__init__(name=name)
        self.create_collection('subspaces')
    
    def create_subspace(self, name, **kwargs):
        pass

    def create_collection(self, name, **kwargs):
        return self._db.get_or_create_collection(name=name, **kwargs)

    def collection(self, name):
        db = self._db.get_collection(name=name)
        return Collection.load(db)
    
    @property
    def collections(self):
        return self._db.get_collections()
    
    @property
    def subspaces(self):
        return self.collection('subspaces')


class App(Space):

    def __init__(self, name):
        super().__init__(name=name)
        self.create_collection('templates')
        self.create_collection('systems')
        self.create_collection('sessions')

    def create_session(self, name, **kwargs):
        pass

    def create_system(self, name, **kwargs):
        pass

    @property
    def templates(self):
        return self.collection('templates')
    
    @property
    def systems(self):
        return self.collection('systems')
    
    @property
    def sessions(self):
        return self.collection('sessions')
    
    def run(self):
        pass


import pytest
import openai
import json

from promptx.template import *
from promptx.models import Response

from . import User, Trait, session, llm


@pytest.fixture
def template(session):
    t = Template(instructions='Some example instructions', output=json.dumps(User.model_json_schema()))
    session.store(t, collection='templates')
    return session.query(ids=[t.id], collection='templates').first


def test_basic_response(session, llm):
    template = Template(instructions='Some example instructions')
    o = session.prompt(template=template, llm=llm)
    assert o is not None
    assert o == 'This is a mock response.'

def test_json_valid_output(session, template, llm):
    llm.generate.return_value = Response(
        raw='{ "name": "test", "age": 20, "traits": ["nice"] }',
    )
    o = session.prompt(template=template, llm=llm)

    assert o is not None
    assert o.type == 'user'
    assert o.name == 'test'
    assert o.age == 20

def test_load_from_template_id(session, template, llm):
    llm.generate.return_value = Response(
        raw='{ "name": "test", "age": 20, "traits": ["nice"] }',
    )
    o = session.prompt(template=template.id, llm=llm)
    assert o is not None
    assert o.type == 'user'
    assert o.name == 'test'
    assert o.age == 20
    
def test_json_valid_output__extra_field(session, template, llm):
    response = Response(
        raw='{ "name": "test", "age": 20, "location": "london", "traits": ["nice"] }',
    )
    llm.generate.return_value = response

    o = session.prompt(template=template, llm=llm)
    assert o.name == 'test'
    assert o.age == 20
    with pytest.raises(AttributeError):
        assert o.location == 'london'
    
def test_json_invalid_output__missing_required_field(session, template, llm):
    response = Response(
        raw='{ "age": 20 }',
    )
    llm.generate.return_value = response

    with pytest.raises(MaxRetriesExceeded):
        session.prompt(template=template, llm=llm)
    
def test_json_invalid_output__formatting(session, template, llm):
    response = Response(
        raw='"name": "test", "age": 20, "traits": ["nice"] }',
    )
    llm.generate.return_value = response

    with pytest.raises(MaxRetriesExceeded):
        session.prompt(template=template, llm=llm)
    
def test_invalild_json_output__validation(session, template, llm):
    response = Response(
        raw='{ "name": "test", "age": "young" }',
    )
    llm.generate.return_value = response

    with pytest.raises(MaxRetriesExceeded):
        session.prompt(template=template, llm=llm)

# TODO: this should probably have some kind of separate retry budget
def test_exception_handling(session, template, llm):
    llm.generate.side_effect = [openai.error.Timeout, Response(raw='Test response')]
    template = Template(instructions='Some example instructions')
    
    o = session.prompt(template=template, llm=llm)
    assert o == 'Test response'

def test_parse_exception_handling(session, mocker, template, llm):
    mocker.patch.object(TemplateRunner, 'process', side_effect=[*[json.JSONDecodeError('test', 'test', 0)] * 4, 'test'])
    runner = TemplateRunner()

    with pytest.raises(MaxRetriesExceeded):
        o = runner(session, template, None, llm=llm)
    
    mocker.patch.object(TemplateRunner, 'process', side_effect=[*[json.JSONDecodeError('test', 'test', 0)] * 3, 'test'])
    runner = TemplateRunner()
    o = runner(session, template, None, llm=llm)
    
    assert o.content == 'test'

def test_invalid_input_raises_error(session, template, llm):
    with pytest.raises(MaxRetriesExceeded):
        session.prompt(template=template, input={'age': 'young'}, llm=llm)

def test_output_parsing(session, template, llm):
    llm.generate.return_value = Response(raw='{ "name": "test", "age": 20, "traits": ["nice"] }')

    o = session.prompt(template=template, llm=llm)
    assert o.type == 'user'
    assert o.name == 'test'
    assert o.age == 20

def test_format_rendering(template):
    runner = TemplateRunner()
    p = runner.render(template, {})
    assert template.instructions in p

def test_format_rendering_with_input(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'Some test input' in p

def test_format_rendering_with_output(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'name (type: string, required: True, default: None' in p

def test_format_rendering_object(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'Return the output as a valid JSON object with the fields described below' in p

def test_format_rendering_list():
    schema = json.dumps({
        'type': 'array',
        'items': {}
    })
    t = Template(instructions='Some example instructions', output=schema)
    runner = TemplateRunner()
    p = runner.render(t, {'input': 'Some test input'})
    assert 'Return a list of valid JSON objects with the fields described below' in p

def test_format_rendering_with_basic_types(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'name (type: string, required: True, default: None' in p
    assert 'age (type: integer, required: True, default: None' in p

def test_format_rendering_with_enum(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'role (type: string, required: False, default: admin' in p
    assert 'Select one option from: admin, user' in p

def test_format_rendering_with_enum_list(session, template, llm):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'traits (type: string[], required: True, default: None' in p
    assert 'Select any relevant options from: nice, mean, funny, smart' in p

def test_format_rendering_with_excluded_fields(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'banned (type: bool, required: False, default: False' not in p

@pytest.mark.skip(reason="Not implemented yet")
def test_format_rendering_with_field_description(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'What kind of personality describes the user?' in p

def test_format_rendering_with_field_min_max(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'minimum: 18' in p
    assert 'exclusiveMaximum: 100' in p

def test_format_rendering_with_field_min_max_items(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'minItems: 1' in p
    assert 'maxItems: 3' in p

def test_format_rendering_with_field_min_max_length(template):
    runner = TemplateRunner()
    p = runner.render(template, {'input': 'Some test input'})
    assert 'minLength: 3' in p
    assert 'maxLength: 20' in p

def test_example_rendering(session):
    user = User(name="John Wayne", age=64, traits=[Trait.mean])
    runner = TemplateRunner()
    template = Template(instructions='Some example instructions', output=user.model_dump_json(),
                        examples=[Example(input='Some test input', output=user.model_dump_json())])
    session.store(template)
    template_r = session.query(ids=[template.id]).first
    p = runner.render(template_r, {'input': 'Some test input'})

    assert 'EXAMPLES' in p
    assert 'John Wayne' in p
    assert '64' in p
    assert 'mean' in p
    assert 'banned' in p

def test_example_rendering_multiple(session, template):
    user = User(name="John Wayne", age=64, traits=[Trait.mean])
    runner = TemplateRunner()
    examples = [Example(input='Some test input', output=user.model_dump_json()) for _ in range(3)]
    template = Template(instructions='Some example instructions', output=user.model_dump_json(), examples=examples)
    session.store(template)
    template = session.query(ids=[template.id]).first
    p = runner.render(template, {'input': 'Some test input'})

    assert p.count('John Wayne') == 3


import os
import click
from rich import pretty, print
import requests
from IPython import embed
from jinja2 import Environment, FileSystemLoader, select_autoescape

from . import load


API_ENDPOINT = os.environ.get('API_ENDPOINT', 'http://localhost:8000')

pretty.install()

@click.group()
def cli():
    pass


@cli.command()
@click.argument('input', nargs=1)
def prompt(input: str):
    _prompt(input)

def _prompt(input: str):
    return requests.post(f'{API_ENDPOINT}/prompt', json={'name': 'n/a', 'instructions': input})


@cli.command(name='query')
@click.argument('texts', nargs=-1)
@click.option('--where', default=None)
@click.option('--field', default=None)
def query(*texts, where=None, field=None):
    _query(*texts, where=where, field=field)

def _query(*texts, where=None, field=None):
    query = { 'texts': texts, 'where': where, 'field': field }
    return requests.post(f'{API_ENDPOINT}/query', data=query)


@cli.command(name='api')
@click.option('--path', default='local')
@click.option('--host', default='0.0.0.0')
@click.option('--port', default='8000')
def api(path=None, host=None, port=None):
    app = load(path)
    import uvicorn
    uvicorn.run(app.api.fastapi_app, host=host, port=port)


@cli.command(name='admin')
@click.option('--path', default='local')
@click.option('--host', default='0.0.0.0')
@click.option('--port', default='8000')
def admin(path=None, host=None, port=None):
    app = load(path)
    from waitress import serve
    serve(app.admin.app.server, host=host, port=port)


@cli.command(name='init')
@click.argument('path', nargs=1, required=False)
def init(path=None):
    # create a hidden directory in the current directory called .px
    if path is None:
        path = os.getcwd()
    elif not os.path.exists(path):
        os.mkdir(path)
    dir = os.path.join(path, '.px')
    if os.path.exists(dir):
        print(f'Error: {dir} already exists')
        return
    os.mkdir(dir)
    print(f'Promptx project created at: {path}')


def main():
    cli()


import json
import logging
import textwrap
import colorama

colorama.just_fix_windows_console()


FORMAT_INSTRUCTIONS = logging.DEBUG + 1
CONTEXT = logging.DEBUG + 2
EXAMPLES = logging.DEBUG + 3
METRICS = logging.DEBUG + 4
HISTORY = logging.DEBUG + 5
INSTRUCTIONS = logging.INFO + 1
INPUT = logging.INFO + 2
OUTPUT = logging.INFO + 3

logging.addLevelName(FORMAT_INSTRUCTIONS, "FORMAT INSTRUCTIONS")
logging.addLevelName(CONTEXT, "CONTEXT")
logging.addLevelName(EXAMPLES, "EXAMPLES")
logging.addLevelName(INSTRUCTIONS, "INSTRUCTIONS")
logging.addLevelName(INPUT, "INPUT")
logging.addLevelName(OUTPUT, "OUTPUT")
logging.addLevelName(METRICS, "METRICS")
logging.addLevelName(HISTORY, "HISTORY")

class NotebookFormatter(logging.Formatter):

    colors = {
        FORMAT_INSTRUCTIONS: colorama.Fore.CYAN,
        CONTEXT: colorama.Fore.BLACK + colorama.Back.YELLOW,
        EXAMPLES: colorama.Fore.MAGENTA,
        METRICS: colorama.Fore.WHITE + colorama.Style.DIM,
        HISTORY: colorama.Back.WHITE + colorama.Fore.BLACK,
        INSTRUCTIONS: colorama.Fore.YELLOW,
        INPUT: colorama.Fore.BLUE,
        OUTPUT: colorama.Fore.GREEN,
        logging.ERROR: colorama.Fore.RED,
        logging.INFO: colorama.Fore.WHITE,
        logging.DEBUG: colorama.Style.DIM,
    }

    def __init__(self, width=80, **kwargs):
        super().__init__(**kwargs)
        self.width = width

    def format(self, record) -> str:
        color = self.colors.get(record.levelno, None)
        msg = f'{color}{record.getMessage()}{colorama.Style.RESET_ALL}'
        return '\n'.join([
            textwrap.fill(line.strip(), width=self.width).ljust(self.width, ' ')
            for line in msg.split('\n')
        ])


class JSONLogFormatter(logging.Formatter):
    def format(self, record):
        msg = super().format(record)
        log_entry = {
            "timestamp": record.created,
            "level": record.levelname,
            "message": msg,
            "filename": record.filename,
            "lineno": record.lineno,
            "funcName": record.funcName,
        }
        return json.dumps(log_entry)


import time
import random
import json
import urllib3
from typing import * 
import jsonschema
from loguru import logger
from pydantic import BaseModel
import openai
from jinja2 import Template as JinjaTemplate

from .collection import Collection, Entity, Query, model_to_json_schema, create_entity_from_schema, serializer, REGISTERED_ENTITIES
from .logging import *
from .models import MockLLM


class MaxRetriesExceeded(Exception):
    pass


E = TypeVar('E', bound=BaseModel)


class Example(Entity):
    input: str
    output: str

    def __init__(self, input, output, **kwargs):
        super().__init__(input=self.parse(input), output=self.parse(output), **kwargs)
    
    def parse(self, x):
        def _serialize(x):
            if issubclass(type(x), BaseModel):
                return x.model_dump()
        if isinstance(x, str):
            return x
        else:
            return json.dumps(x, default=_serialize)


class Template(Entity):
    
    template: str = """
    INSTRUCTIONS
    ---
    {{instructions}}
    {{format}}
    {% if examples %}
        EXAMPLES
        ---
    {% endif %}
    {{examples}}
    {% if examples %}
        END_EXAMPLES
        ---
    {% endif %}
    {{input}}
    {{output}}
    """

    input_template: str = """
    INPUT
    ---
    {{input}}
    END_INPUT
    """

    output_template: str = """
    OUTPUT
    ---
    {{output}}
    """

    example_template: str = f"""
    {input_template}
    {output_template}
    """

    format_template: str = """
    FORMAT INSTRUCTIONS
    ---
    {% if string_list_output %}
    Return a JSON array of strings.
    {% elif list_output %}
    Return a list of valid JSON objects with the fields described below.
    {% else %}
    Return the output as a valid JSON object with the fields described below. 
    {% endif %}
    {% for field in fields %}
    - {{field.name}} (type: {{field.type_}}, required: {{field.required}}, default: {{field.default}}, {% for k, v in field.metadata.items()%}{{k}}: {{v}}, {% endfor %}): {{field.description}}
    {% endfor %}

    Make sure to use double quotes and avoid trailing commas!
    Ensure any required fields are set, but you can use the default value 
    if it's defined and you are unsure what to use. 
    If you are unsure about any optional fields use `null` or the default value,
    but try your best to fill them out.
    END_FORMAT_INSTRUCTIONS
    """

    type: str = 'template'
    name: str = None
    instructions: str = None
    examples: List[Example] = None
    input: str = None
    output: str = None
    context: str = None
    data: Query = None

    def __init__(self, examples=None, **kwargs):
        if examples is not None:
            for i, example in enumerate(examples):
                if isinstance(example, dict):
                    example = Example(**example)
                elif isinstance(example, tuple):
                    example = Example(*example)
                if not isinstance(example, Example):
                    continue
                examples[i] = example
            kwargs['examples'] = examples
        super().__init__(**kwargs)


class TemplateRunner:
    
    def parse(self, x):
        if x is None:
            return {}
        skip_list = ['id', 'type']
        if isinstance(x, BaseModel):
            skip_list += [k for k, v in x.model_fields.items() if v.json_schema_extra and v.json_schema_extra.get('generate') == False]
            return {k: v for k, v in x.model_dump().items() if k not in skip_list}
        elif isinstance(x, Entity):
            skip_list += [k for k, v in x.model_fields.items() if v.json_schema_extra and v.json_schema_extra.get('generate') == False]
            return {k: v for k, v in x.object.dict().items() if k not in skip_list}
        elif isinstance(x, Collection):
            return [
                {k: v for k, v in y.dict().items() if k not in skip_list}
                for y in x.objects
            ]
        elif isinstance(x, str):
            return x
        elif isinstance(x, dict):
            return {k: self.parse(v) for k, v in x.items()}
        elif isinstance(x, list):
            return [self.parse(y) for y in x]
        else:
            return x
    
    def render(self, t, x, **kwargs):
        input_template = JinjaTemplate(t.input_template)
        input = input_template.render(**x) if len(x) > 0 else ''
        output_template = JinjaTemplate(t.output_template)
        output = output_template.render(**x) if len(x) > 0 else ''
        vars = {
            **x,
            'instructions': t.instructions,
            'examples': self.render_examples(t),
            'format': self.render_format(t, x),
            'input': input,
            'output': output,
        }
        template = JinjaTemplate(t.template)
        output = template.render(**vars)
        return output
    
    def format_field(self, name, field, definitions, required):
        if name in ['id', 'type']:
            return None
        if field.get('generate', True) == False:
            return None
        description = field.get('description', '')
        options = ''
        metadata_keys = [
            'minimum', 'maximum', 'exclusiveMinimum', 'exclusiveMaximum',
            'minLength', 'maxLength',
            'minItems', 'maxItems',
        ]
        metadata = {
            k: v for k, v in field.items()
            if k in metadata_keys
        }

        definition = None
        list_field = False
        if field.get('type') == 'array':
            list_field = True
            item_type = field.get('items', {}).get('type', None)
            if item_type is None:
                ref = field.get('items', {}).get('$ref', None)
                ref = ref.split('/')[-1]
                definition = definitions.get(ref, {})
                type_ = f'{definition.get("type")}[]'
            else:
                type_ = f'{item_type}[]'
            field = field.get('items', {})
        elif len(field.get('allOf', [])) > 0:
            ref = field.get('allOf')[0].get('$ref')
            ref = ref.split('/')[-1]
            definition = definitions.get(ref, {})
            type_ = f'{definition.get("type")}'
        elif field.get('$ref'):
            ref = field.get('$ref')
            ref = ref.split('/')[-1]
            definition = definitions.get(ref, {})
            type_ = f'{definition.get("type")}'
        else:
            type_ = field.get('type', 'str')
        
        if definition is not None and 'enum' in definition:
            if list_field:
                options += f'''
                Select any relevant options from: {", ".join(definition["enum"])}
                '''
            else:
                options += f'''
                Select one option from: {", ".join(definition["enum"])}
                '''

        if len(options) > 0:
            description += ' ' + options

        return {
            'name': name,
            'title': field.get('title', None),
            'type_': type_,
            'default': field.get('default', None),
            'description': description.strip(),
            'required': name in required,
            'metadata': metadata,
        }
    
    def render_format(self, t, x, **kwargs):
        if t.output is None or t.output == str:
            return ''
        
        output = json.loads(t.output)
        format_template = JinjaTemplate(t.format_template)
        if output.get('type', None) == 'array' and output.get('items', {}).get('type', None) == 'string':
            return format_template.render({
                'string_list_output': True,
            })

        list_output = False
        fields = []
        properties = {}
        if output.get('type', None) == 'array':
            properties = output.get('items', {}).get('properties', {})
            definitions = output.get('items', {}).get('$defs', {})
            required = output.get('items', {}).get('required', [])
            list_output = True
        elif output.get('type', None) == 'object':
            properties = output.get('properties', {})
            definitions = output.get('$defs', {})
            required = output.get('required', [])
        
        for name, property in properties.items():
            f = self.format_field(name, property, definitions, required)
            fields += [f]
        
        return format_template.render({
            'fields': [field for field in fields if field is not None], 
            'list_output': list_output,
        })
    
    def render_examples(self, t, **kwargs):
        if t.examples is None or len(t.examples) == 0:
            return ''
        
        examples = [
            {
                'input': e.input,
                'output': e.output,
            }
            for e in random.sample(t.examples, min(len(t.examples), 3))
        ]
        example_template = JinjaTemplate(t.example_template)
        return '\n'.join([
            example_template.render(**e) for e in examples
        ])
    
    def process(self, session, t, x, output, allow_none=False, **kwargs):
        if t.output is None:
            return output
        if allow_none and output is None:
            return None
        out = json.loads(output)
        schema = model_to_json_schema(json.loads(t.output))
        if schema.get('type', None) == 'string' or (schema.get('type', None) == 'array' and schema.get('items', {}).get('type', None) == 'string'):
            return out
        entities = create_entity_from_schema(schema, out, session=session, base=Entity)
        return entities
    
    def dict(self):
        return {
            'id': self.id,
            'type': 'template',
            'name': self.name or None,
            'instructions': self.instructions,
            'input': self.input,
            'output': self.output,
        }
    
    def __call__(self, session, t, x, llm, **kwargs):
        return self.forward(session, t, x, llm, **kwargs)
    
    def forward(self, session, t, x, llm, context=None, history=None, retries=3, dryrun=False, allow_none=False, **kwargs):
        if retries and retries <= 0:
            e = MaxRetriesExceeded(f'{t.name} failed to forward {x}')
            logger.error(e)
            raise e
        
        if dryrun:
            logger.debug(f'Dryrun: {t.output}')
            llm = MockLLM(output=t.output)
        
        px = self.parse(x)
            
        prompt_input = self.render(t, {'input': px})

        try:
            response = llm.generate(prompt_input, context=context or t.context, history=history)
        except (urllib3.exceptions.ReadTimeoutError,
                urllib3.exceptions.ConnectTimeoutError,
                urllib3.exceptions.NewConnectionError) as e:
            logger.error(f'LLM generation failed: {e}')
            time.sleep(2)
            return self.forward(session, t, x, llm, retries=retries, **kwargs)
        except (openai.error.APIError,
                openai.error.Timeout,
                openai.error.ServiceUnavailableError) as e:
            logger.error(f'LLM generation failed: {e}')
            time.sleep(2)
            return self.forward(session, t, x, llm, retries=retries, **kwargs)
        except openai.error.RateLimitError as e:
            logger.error(f'Hit rate limit for {self}: {e}')
            time.sleep(10)
            return self.forward(session, t, x, llm, retries=retries, **kwargs)

        try:
            response.content = self.process(session, t, px, response.raw, allow_none=allow_none, **kwargs)
        except jsonschema.exceptions.ValidationError as e:
            logger.error(f'Output validation failed: {e}')
            return self.forward(session, t, x, llm, retries=retries-1, **kwargs)
        except json.JSONDecodeError as e:
            logger.warning(f'Failed to decode JSON from {e}')
            return self.forward(session, t, x, llm, retries=retries-1, **kwargs)
        except Exception as e:
            logger.error(f'Failed to forward {x}: {e}')
            return self.forward(session, t, x, llm, retries=retries-1, **kwargs)
        return response
